import os

from openai import OpenAI

from rag4p.rag.generation.chat.chat_prompt import ChatPrompt
from rag4p.rag.generation.quality.answer_from_context_quality import AnswerFromContextQuality
from rag4p.rag.generation.quality.answer_quality import AnswerQuality
from rag4p.rag.generation.quality.answer_to_question_quality import AnswerToQuestionQuality
from rag4p.rag.tracker.rag_observer import RAGObserver


class AnswerQualityService:
    def __init__(self, openai_api_key: str):
        self.openai_client = OpenAI(
            api_key=openai_api_key,
        )

    def determine_quality_of_answer(self, rag_observer: RAGObserver) -> AnswerQuality:
        answer_from_context = self.determine_quality_answer_from_context(rag_observer)
        answer_to_question = self.determine_quality_answer_related_to_question(rag_observer)

        return AnswerQuality(answer_to_question, answer_from_context)

    def determine_quality_answer_related_to_question(self, rag_observer: RAGObserver) -> AnswerToQuestionQuality:
        chat_prompt = ChatPrompt(system_message_filename="../data/quality/quality_of_answer_to_question_system.txt",
                                 user_message_filename="../data/quality/quality_of_answer_to_question_user.txt")

        completion = self.openai_client.chat.completions.create(
            model="gpt-4",
            messages=[
                {"role": "system",
                 "content": chat_prompt.create_system_message(params={})},
                {"role": "user",
                 "content": chat_prompt.create_user_message(params={
                     "question": rag_observer.question,
                     "answer": rag_observer.answer})},
            ],
            stream=False,
        )
        quality, reason = self.split_quality_and_reason(completion.choices[0].message.content)
        return AnswerToQuestionQuality(quality, reason)

    def determine_quality_answer_from_context(self, rag_observer: RAGObserver) -> AnswerFromContextQuality:
        chat_prompt = ChatPrompt(system_message_filename="../data/quality/quality_of_answer_from_context_system.txt",
                                 user_message_filename="../data/quality/quality_of_answer_from_context_user.txt")

        completion = self.openai_client.chat.completions.create(
            model="gpt-4",
            messages=[
                {"role": "system",
                 "content": chat_prompt.create_system_message(params={})},
                {"role": "user",
                 "content": chat_prompt.create_user_message(params={
                     "context": rag_observer.context,
                     "answer": rag_observer.answer})},
            ],
            stream=False,
        )
        quality, reason = self.split_quality_and_reason(completion.choices[0].message.content)
        return AnswerFromContextQuality(quality, reason)

    def split_quality_and_reason(self, quality_and_reason: str) -> (int, str):
        split = quality_and_reason.split("-", 1)
        quality = int(split[0].strip())
        reason = split[1].strip()
        return quality, reason


from typing import List

import pandas as pd

from rag4p.rag.model.chunk import Chunk
from rag4p.rag.store.content_store import ContentStore
from rag4p.rag.embedding.embedder import Embedder
from rag4p.rag.retrieval.retriever import Retriever

from scipy.spatial import distance


class InternalContentStore(ContentStore, Retriever):
    """
    The internal content stores stores the chunks in memory, it acts as a normal content store, but it als contains
    all the methods from a retriever, so it can be used as a retriever as well. This is useful for testing purposes.
    """
    def __init__(self, embedder: Embedder):
        self.embedder = embedder
        self.vector_store = pd.DataFrame(columns=['chunk_id', 'chunk', 'embedding'])

    def store(self, chunks: List[Chunk]):
        for chunk in chunks:
            self.__store_chunk(chunk)

    def __store_chunk(self, chunk: Chunk):
        chunk_id = chunk.document_id + "_" + str(chunk.chunk_id)
        print(f"Storing chunk {chunk_id}: {chunk.chunk_text}")
        embedding = self.embedder.embed(chunk.chunk_text)
        self.vector_store.loc[len(self.vector_store)] = {'chunk_id': chunk_id, 'chunk': chunk, 'embedding': embedding}

    def find_relevant_chunks(self, query: str, max_results: int = 4) -> List[Chunk]:
        print(f"Finding relevant chunks for query: {query}")
        embedding = self.embedder.embed(query)
        self.vector_store['distance'] = self.vector_store['embedding'].apply(lambda x: distance.euclidean(x, embedding))
        relevant_chunks_df = self.vector_store.nsmallest(max_results, 'distance')
        return relevant_chunks_df['chunk'].tolist()

    def get_chunk_by_id(self, chunk_id: str) -> Chunk:
        """
        Obtains a chunk using its complete id (document_id + "_" + chunk_id)
        :param chunk_id: complete id of the chunk document_id + "_" + chunk_id
        :return:
        """
        found_chunk = self.vector_store[self.vector_store['chunk_id'] == chunk_id]

        if len(found_chunk) == 0:
            raise Exception(f"Chunk with id {chunk_id} not found.")

        return found_chunk.iloc[0]['chunk']

    def loop_over_chunks(self):
        for index, row in self.vector_store.iterrows():
            yield row['chunk']


import json

from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.content_reader import ContentReader


class VasaContentReader(ContentReader):
    def __init__(self):
        self.filename = "../data/vasa-timeline.jsonl"

    def read(self):
        with open(self.filename, 'r') as file:
            for line in file:
                data = json.loads(line)
                properties = {
                    "title": data["title"],
                    "timerange": data["timerange"]
                }
                document_id = data["title"].lower().replace(" ", "-")
                document = InputDocument(
                    document_id=document_id,
                    text=data["body"],
                    properties=properties
                )
                yield document



from abc import ABC, abstractmethod
from typing import List

from rag4p.rag.model.chunk import Chunk


class ContentStore(ABC):

    @abstractmethod
    def store(self, chunks: List[Chunk]):
        pass
